import os

from django.conf import settings

from txn_notifications.templates import BASE_DIR

DEFAULTS = {
    "django": {
        "enabled": False,
    },
    "email": {
        "enabled": False,
        # default options
        "html_template": os.path.join(
            BASE_DIR, "txn_notifications/templates/email.html"
        ),
        "txt_template": os.path.join(
            BASE_DIR, "txn_notifications/templates/email.txt"
        ),
        # how to get the recipient
        "recipient_attr": "email",
    },
    "twilio": {
        # twilio handler is enabled
        "enabled": False,
        # account settings
        "account_sid": None,
        "auth_token": None,
        "sms_sender": None,
        # how to get the recipient
        "recipient_attr": None,
    },
}


def get_settings():
    user_settings = settings.DJANGO_NOTIFICATIONS
    custom_settings = DEFAULTS.copy()
    custom_settings.update(user_settings)
    return custom_settings



from django.conf import settings
from django.utils import timezone

from swapper import load_model
from twilio.rest import Client

from txn_notifications.exceptions import NotificationNotSent

from .generic import Handler

Notification = load_model("notification", "Notification")


class SMS(Handler):
    def __init__(self):
        super().__init__()


class SMSTwilioHandler(SMS):
    slug = "twilio"
    recipient_attr = None

    enabled = False
    sms_sender = None

    def __init__(self, *args, **kwargs):
        twilio = settings.DJANGO_NOTIFICATIONS.get("twilio")

        if not twilio.get("enabled", self.enabled):
            raise NotificationNotSent("twilio is not enabled.")

        self.recipient_attr = twilio.get("recipient_attr", self.recipient_attr)

        self.callback = twilio.get("callback", self.get_callback())

        self.__sms_sender = twilio.get("sms_sender", self.sms_sender)

        self.__client = Client(
            twilio.get("account_sid"), twilio.get("auth_token")
        )
        super().__init__(*args, **kwargs)

    def perform_send(self):
        sms_msg = self.get_sms_message()

        if self.template.check_status:
            if not self.callback:
                raise NotificationNotSent("no callback")

            # send notification to twilio (with callback)
            self.notification = self.__client.messages.create(
                to=self.recipient,
                from_=self.__sms_sender,
                body=sms_msg,
                status_callback=self.callback,
            )
        else:
            # send notification to twilio
            self.notification = self.__client.messages.create(
                to=self.recipient,
                from_=self.__sms_sender,
                body=sms_msg,
            )

    def get_sms_message(self):
        sms_msg = f"{self.title}"

        if self.body:
            sms_msg = f"{sms_msg} {self.body}"

        if self.url:
            if self.url_msg:
                sms_msg = f"{sms_msg} {self.url_msg} {self.url}"
            else:
                sms_msg = f"{sms_msg} {self.url}"

        return sms_msg

    @classmethod
    def get_prov_id(cls, response) -> str:
        """
        See Also:
            https://www.twilio.com/docs/sms/send-messages
        Args:
            response: Expected response when a message is sent:
                {
                  "account_sid": "ACXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "api_version": "2010-04-01",
                  "body": "McAvoy or Stewart? These timelines can get so
                    confusing.",
                  "date_created": "Thu, 30 Jul 2015 20:12:31 +0000",
                  "date_sent": "Thu, 30 Jul 2015 20:12:33 +0000",
                  "date_updated": "Thu, 30 Jul 2015 20:12:33 +0000",
                  "direction": "outbound-api",
                  "error_code": null,
                  "error_message": null,
                  "from": "+15017122661",
                  "messaging_service_sid":
                    "MGXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "num_media": "0",
                  "num_segments": "1",
                  "price": null,
                  "price_unit": null,
                  "sid": "SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "status": "sent",
                  "subresource_uris": {
                    "media": "/2010-04-01/Accounts/ACXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXXXXX/Messages/SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    /Media.json"
                  },
                  "to": "+15558675310",
                  "uri": "/2010-04-01/Accounts/ACXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXX/Messages/SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX.json"
                }

        Returns:
            (str): response.data.sid
        """
        data = response.json()
        return data.get("sid")

    @classmethod
    def get_prov_status(cls, response):
        """
        See Also:
            https://www.twilio.com/docs/sms/send-messages
            https://www.twilio.com/docs/sms/api/message-resource#message-status-values
        Args:
            response: Expected response when a message is sent:
                {
                  "account_sid": "ACXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "api_version": "2010-04-01",
                  "body": "McAvoy or Stewart? These timelines can get so
                    confusing.",
                  "date_created": "Thu, 30 Jul 2015 20:12:31 +0000",
                  "date_sent": "Thu, 30 Jul 2015 20:12:33 +0000",
                  "date_updated": "Thu, 30 Jul 2015 20:12:33 +0000",
                  "direction": "outbound-api",
                  "error_code": null,
                  "error_message": null,
                  "from": "+15017122661",
                  "messaging_service_sid":
                    "MGXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "num_media": "0",
                  "num_segments": "1",
                  "price": null,
                  "price_unit": null,
                  "sid": "SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "status": "sent",
                  "subresource_uris": {
                    "media": "/2010-04-01/Accounts/ACXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXXXXX/Messages/SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    /Media.json"
                  },
                  "to": "+15558675310",
                  "uri": "/2010-04-01/Accounts/ACXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXX/Messages/SMXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX.json"
                }

        Returns:
            ([str, str]): response.data.status, response.data.error_code
        """
        data = response.json()
        return data.get("status"), data.get("error_code")

    @classmethod
    def update_status(cls, response, callback_id=None):
        """

        Args:
            response:
            callback_id:

        See Also:
            # https://www.twilio.com/docs/sms/api/message-resource#message-status-values

        Returns:

        """
        notification = Notification.objects.get(callback_id=callback_id)

        status, error_code = cls.get_prov_status(response)

        if not notification.is_sent:
            notification.is_sent = status == "sent"
            if notification.is_sent:
                notification.sent_timestamp = timezone.now()

        if error_code:
            notification.data["error_code"] = error_code

        notification.status = status
        notification.save()



from django.test import TestCase

from txn_notifications.tests.factories import UserFactory, TemplateFactory


class TemplateTestCase(TestCase):
    def setUp(self):
        self.recipient = UserFactory()
        self.template = TemplateFactory(
            title="{{ recipient.first_name }} **markdown**",
            body="{{ recipient.first_name }} **markdown**",
            url="{{ recipient.first_name }} **markdown**",
            url_msg="{{ recipient.first_name }} **markdown**",
        )



"""
Django settings for xxx project.

Generated by 'django-admin startproject' using Django 4.0.6.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""
import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = (
    "django-insecure-wcpjc5!7frb0snk5ow$ahkyz8o1=4425*$qn80b5bvc7+sfe3v"
)

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "txn_notifications",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "txn_notifications.config.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation."
        "UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation."
        "MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation."
        "CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation."
        "NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# for testing
DJANGO_NOTIFICATIONS = {"test": {"enbled": True, "recipient_attr": "email"}}

# email
DEFAULT_FROM_EMAIL = "Test <test@test.com>"
EMAIL_BACKEND = "django.core.mail.backends.console.EmailBackend"

if os.environ.get("SAMPLE", False):
    INSTALLED_APPS += ["txn_notifications.tests.sample_notifications"]

    TXN_NOTIFICATIONS_CATEGORY_MODEL = "sample_notifications.Category"
    TXN_NOTIFICATIONS_TEMPLATE_MODEL = "sample_notifications.Template"
    TXN_NOTIFICATIONS_NOTIFICATION_MODEL = "sample_notifications.Notification"
    TXN_NOTIFICATIONS_USERHANDLERSETTING_MODEL = (
        "sample_notifications.UserHandlerSetting"
    )
    TXN_NOTIFICATIONS_USERCATEGORYSETTING_MODEL = (
        "sample_notifications.UserCategorySetting"
    )

    TEMPLATES[0]["DIRS"] += [os.path.join(BASE_DIR, "../templates")]



from django.utils import timezone

from swapper import load_model

from .generic import Handler


class DjangoHandler(Handler):
    slug = "django"
    recipient_attr = None  # uses self.user

    def send(self):
        Notification = load_model("txn_notifications", "Notification")

        recipient = self.get_recipient()
        self.notification = Notification.objects.create(
            recipient=recipient,
            sender=self.sender,
            template=self.template,
            title=self.title,
            body=self.body,
            url=self.url,
            url_msg=self.url_msg,
            sent=True,
            sent_timestamp=timezone.now(),
        )

    def post_send(self, response):
        pass  # recorded on send()

    def record_notification(self):
        pass  # recorded on send()

    def get_recipient(self):
        return self.user



import re

from django import template
from django.template.defaultfilters import stringfilter

import markdown as md
from bs4 import BeautifulSoup

register = template.Library()


@register.filter
@stringfilter
def markdown(value):
    return md.markdown(value, extensions=["markdown.extensions.fenced_code"])


@register.filter
@stringfilter
def simple_text(value):
    """Converts a markdown string to plaintext"""

    # md -> html -> text since BeautifulSoup can extract text cleanly
    html = markdown(value)

    # remove code snippets
    html = re.sub(r"<pre>(.*?)</pre>", " ", html)
    html = re.sub(r"<code>(.*?)</code >", " ", html)

    # extract text
    soup = BeautifulSoup(html, "html.parser")
    text = "".join(soup.findAll(string=True))

    return text



from django.test import TestCase
from django.core.exceptions import ObjectDoesNotExist

import factory
from swapper import load_model

from txn_notifications.wrappers import NotificationWrapper
from txn_notifications.exceptions import (
    WrapperError,
    HandlerDisallowed,
    CategoryDisallowed,
    NotificationNotSent,
)
from txn_notifications.handlers.django import DjangoHandler
from txn_notifications.tests.factories import (
    UserFactory,
    CategoryFactory,
    TemplateFactory,
    UserHandlerSettingFactory,
    UserCategorySettingFactory,
)

Template = load_model("txn_notifications", "Template")


class NotificationWrapperTestCase(TestCase):
    def setUp(self):
        self.user = UserFactory(email=factory.Faker("email"))

    def setup_min_wrapper(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            title = "title"

            supported_handlers = [DjangoHandler]

        return Wrapper

    # ==========================================
    # send
    # ==========================================
    def test_send__min(self):
        Wrapper = self.setup_min_wrapper()

        Wrapper.create_templates()
        self.assertEqual(Template.objects.count(), 1)
        template = Template.objects.first()

        self.assertEqual(self.user.notifications.count(), 0)

        notification = Wrapper()
        notification.send(recipient=self.user)

        self.assertEqual(self.user.notifications.count(), 1)

        # check user notifications
        self.assertEqual(self.user.notifications.count(), 1)
        notification = self.user.notifications.first()

        # status
        self.assertFalse(notification.read)
        self.assertIsNone(notification.read_timestamp)
        self.assertTrue(notification.sent)
        self.assertIsNotNone(notification.sent_timestamp)

        # related
        self.assertIsNone(notification.sender)
        self.assertEqual(notification.template, template)
        self.assertEqual(notification.title, Wrapper.title)
        self.assertEqual(notification.body, "")
        self.assertEqual(notification.url, "")
        self.assertEqual(notification.url_msg, "")

    def test_send__sender(self):
        sender = UserFactory()

        Wrapper = self.setup_min_wrapper()

        Wrapper.create_templates()
        self.assertEqual(Template.objects.count(), 1)
        template = Template.objects.first()

        self.assertEqual(self.user.notifications.count(), 0)

        notification = Wrapper()
        notification.send(recipient=self.user, sender=sender)

        # check user notifications
        self.assertEqual(self.user.notifications.count(), 1)
        notification = self.user.notifications.first()

        # status
        self.assertFalse(notification.read)
        self.assertIsNone(notification.read_timestamp)
        self.assertTrue(notification.sent)
        self.assertIsNotNone(notification.sent_timestamp)

        # related
        self.assertEqual(notification.sender, sender)
        self.assertEqual(notification.template, template)
        self.assertEqual(notification.title, Wrapper.title)
        self.assertEqual(notification.body, "")
        self.assertEqual(notification.url, "")
        self.assertEqual(notification.url_msg, "")

    def test_send__data(self):
        pass

    def test_send__context(self):
        pass

    def test_send__no_templates(self):
        class Wrapper(NotificationWrapper):
            supported_handlers = [DjangoHandler]

        try:
            Wrapper().send(recipient=self.user)
            self.fail("WrapperException must be raised")
        except WrapperError as e:
            self.assertEqual(e.code, "no_active_template")

    # ==========================================
    # data context
    # ==========================================
    def test_get_default_data(self):
        sender = UserFactory()

        Wrapper = self.setup_min_wrapper()
        data = Wrapper.get_data(
            recipient=self.user, sender=sender, target=sender
        )
        self.assertEqual(data.get("recipient"), self.user)
        self.assertEqual(data.get("sender"), sender)
        self.assertEqual(data.get("target"), sender)

    def test_get_data__custom(self):
        sender_user = UserFactory()

        class Wrapper(NotificationWrapper):
            @classmethod
            def get_data(cls, *args, **kwargs):
                data = super().get_data(*args, **kwargs)
                data["foo"] = "bar"
                return data

        data = Wrapper.get_data(recipient=self.user, sender=sender_user)
        self.assertEqual(data.get("recipient"), self.user)
        self.assertEqual(data.get("sender"), sender_user)
        self.assertEqual(data.get("foo"), "bar")

    def test_get_handler(self):
        Wrapper = self.setup_min_wrapper()
        Wrapper.create_templates()
        template = Template.objects.first()

        handler = Wrapper.get_handler(template)
        self.assertEqual(handler, Wrapper.supported_handlers[0])

    def test_get_category(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            title = "title"

            template_category = "category"

        _category = CategoryFactory(name="category", slug="category")

        category = Wrapper._get_category()
        self.assertEqual(category.id, _category.id)
        self.assertEqual(category.slug, _category.slug)
        self.assertEqual(category.name, _category.name)

    def test_get_category__no_template_category(self):
        Wrapper = self.setup_min_wrapper()
        self.assertIsNone(Wrapper._get_category())

    def test_get_category__doesnt_exists(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            title = "title"

            template_category = "x"

        try:
            Wrapper._get_category()
            self.fail("expected ObjectDoesNotExist")
        except ObjectDoesNotExist:
            pass

    # ==========================================
    # templates
    # ==========================================
    def test_create_templates(self):
        Wrapper = self.setup_min_wrapper()

        self.assertEqual(Template.objects.count(), 0)
        Wrapper.create_templates()
        self.assertEqual(Template.objects.count(), 1)

        template = Template.objects.first()
        self.assertEqual(template.name, Wrapper.template_name)
        self.assertEqual(
            template.slug, Wrapper.get_template_slug(DjangoHandler)
        )
        self.assertEqual(template.title, Wrapper.title)
        self.assertEqual(template.handler, DjangoHandler.slug)
        self.assertIsNone(template.category)
        self.assertTrue(template.is_active)

    def test_create_templates__missing_defaults(self):
        class Wrapper(NotificationWrapper):
            supported_handlers = [DjangoHandler]

        try:
            Wrapper.create_templates()
        except WrapperError as e:
            self.assertEqual(e.code, "missing_defaults")

    def test_create_templates__no_handlers(self):
        class Wrapper(NotificationWrapper):
            supported_handlers = []

        try:
            Wrapper()
            self.fail("Wrapper exception must be raised")
        except WrapperError as e:
            self.assertEqual(e.code, "no_handlers")

    def test_create_templates__default_attributes(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            template_description = "description"

            title = "title"
            body = "body"
            url = "url"
            url_msg = "url_msg"
            is_active = True

            supported_handlers = [DjangoHandler]

        self.assertEqual(Template.objects.count(), 0)

        Wrapper.create_templates()

        self.assertEqual(Template.objects.count(), 1)
        template = Template.objects.first()

        self.assertEqual(
            template.slug, Wrapper.get_template_slug(DjangoHandler)
        )
        self.assertEqual(template.name, Wrapper.template_name)
        self.assertEqual(template.description, Wrapper.template_description)
        self.assertEqual(template.title, Wrapper.title)
        self.assertEqual(template.body, Wrapper.body)
        self.assertEqual(template.url, Wrapper.url)
        self.assertEqual(template.url_msg, Wrapper.url_msg)
        self.assertTrue(template.is_active)

    def test_create_templates__custom_attributes(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            template_description = "description"

            title = "title"
            body = "body"
            url = "url"
            url_msg = "url_msg"
            is_active = True

            django_title = "django title"
            django_body = "django body"
            django_url = "django url"
            django_url_msg = "django url msg"
            django_is_active = False

            supported_handlers = [DjangoHandler]

        self.assertEqual(Template.objects.count(), 0)

        Wrapper.create_templates()

        self.assertEqual(Template.objects.count(), 1)
        template = Template.objects.first()

        self.assertEqual(
            template.slug, Wrapper.get_template_slug(DjangoHandler)
        )
        self.assertEqual(template.name, Wrapper.template_name)
        self.assertEqual(template.description, Wrapper.template_description)
        self.assertEqual(template.title, Wrapper.django_title)
        self.assertEqual(template.body, Wrapper.django_body)
        self.assertEqual(template.url, Wrapper.django_url)
        self.assertEqual(template.url_msg, Wrapper.django_url_msg)
        self.assertEqual(template.is_active, Wrapper.django_is_active)

    def test_create_templates__category(self):
        category = CategoryFactory(name="Category name", slug="category")

        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            template_description = "description"
            template_category = category.slug

            title = "title"

            supported_handlers = [DjangoHandler]

        Wrapper.create_templates()

        self.assertEqual(Template.objects.count(), 1)
        template = Template.objects.first()
        self.assertEqual(template.category, category)

    def test_create_templates__no_category(self):
        class Wrapper(NotificationWrapper):
            template_name = "name"
            template_slug = "slug"
            template_description = "description"
            template_category = "xxx"

            title = "title"

            supported_handlers = [DjangoHandler]

        try:
            Wrapper.create_templates()
            self.fail("expected ObjectDoesNotExist")
        except ObjectDoesNotExist:
            pass

    def test_update_templates(self):
        pass

    def test_update_templates__default_attributes(self):
        pass

    def test_update_templates__custom_attributes(self):
        pass

    def test_update_templates__category(self):
        pass

    def test_update_templates__no_category(self):
        pass

    # ==========================================
    # check funcions
    # ==========================================
    def test_check_recipient(self):
        self.assertTrue(self.user.is_active)
        NotificationWrapper.check_recipient(self.user)

    def test_check_recipient__no_active(self):
        self.user.is_active = False
        self.user.save()

        try:
            NotificationWrapper.check_recipient(self.user)
            self.fail("NotificationNotSent should be raised")
        except NotificationNotSent as e:
            self.assertEqual(e.code, "no_active_rcpt")

    def test_check_recipient_settings_handler__allow(self):
        pass

    def test_check_recipient_settings_handler__disallow(self):
        settings = UserHandlerSettingFactory(
            handler=DjangoHandler.slug, allow=False
        )
        user = settings.user

        template = TemplateFactory(handler=DjangoHandler.slug)
        try:
            NotificationWrapper.check_recipient_settings(user, template)
            self.fail("HandlerDisallowed expected")
        except HandlerDisallowed as e:
            self.assertEqual(e.code, "disallowed_handler")

    def test_check_recipient_settings_category__allow(self):
        category = CategoryFactory()

        settings = UserCategorySettingFactory(category=category, allow=True)
        user = settings.user

        template = TemplateFactory(category=category)
        self.assertEqual(settings.category, template.category)

        NotificationWrapper.check_recipient_settings(user, template)

    def test_check_recipient_settings_category__disallow(self):
        category = CategoryFactory()

        settings = UserCategorySettingFactory(category=category, allow=False)
        user = settings.user

        template = TemplateFactory(category=category)
        try:
            NotificationWrapper.check_recipient_settings(user, template)
            self.fail("CategoryDisallowed expected")
        except CategoryDisallowed as e:
            self.assertEqual(e.code, "disallowed_category")



from django.template import Context
from django.template import Template as DjangoTemplate

from swapper import load_model

from txn_notifications.settings import get_settings
from txn_notifications.exceptions import NotificationNotSent

settings = get_settings()


class Handler:
    slug = None
    recipient_attr = None

    check_status = None
    record = None

    def __init__(
        self,
        recipient,
        template,
        *,
        sender=None,
        data=None,
        **kwargs,
    ):
        if not self.slug:
            raise NotificationNotSent("handler must have a slug.")

        if not settings.get(self.slug):
            raise NotificationNotSent(
                f"handler '{self.slug}' setttings are required."
            )

        self.recipient_attr = settings.get(self.slug, {}).get(
            "recipient_attr", self.recipient_attr
        )

        self.user = recipient  # User object
        self.template = template
        self.sender = sender
        self.data = data
        self.context = data.get("context", {})

        # prepare the message to the provider
        self.title = self.get_title()
        self.body = self.get_body()
        self.url = self.get_url()
        self.url_msg = self.get_url_msg()

        self.data["title"] = self.title
        self.data["body"] = self.body
        self.data["url"] = self.url
        self.data["url_msg"] = self.url_msg

        # gets the string recipient to send it to the provider.
        # For example: for an email it migth be user.email
        self.recipient = self.get_recipient()
        self.data["recipient_attr"] = self.recipient

        self.notification = None

    def pre_send(self) -> None:
        """
        Data formating and additional validations.
        """

    def send(self) -> None:
        """
        Prepares, sends and saves the notification info.
        """
        self.pre_send()
        response = self.perform_send()
        self.post_send(response)

    def perform_send(self) -> object:
        """
        Send the notification to the provider.

        Returns:
            Provider's http response
        """
        raise NotImplementedError()

    def post_send(self, response) -> None:

        if self.record is not None and not self.record:
            pass
        elif self.template.record:
            # record on the database
            # populate self.notification
            self.record_notification()

        if self.check_status is not None and not self.check_status:
            pass
        elif self.template.check_status:
            self.notification.prov_id = self.get_prov_id(response)
            self.notification.prov_status = self.get_prov_status(response)
            self.notification.save()

    def record_notification(self):
        Notification = load_model("txn_notifications", "Notification")
        self.notification = Notification.objects.create(
            recipient=self.user,
            sender=self.sender,
            template=self.template,
            title=self.title,
            body=self.body,
            url=self.url,
            url_msg=self.url_msg,
        )

    def get_title(self) -> str:
        """
        Renders the title.
        Returns:
            (str): The title rendered.
        """
        context = Context(self.data)
        template = DjangoTemplate(self.template.title)
        return template.render(context)

    def get_body(self) -> str:
        """
        Renders the body.
        Returns:
            (str): The body rendered or empty string if is None.
        """
        if not self.template.body:
            return ""

        context = Context(self.data)
        template = DjangoTemplate(self.template.body)
        return template.render(context)

    def get_url(self) -> str:
        """
        Renders the url.
        Returns:
            (str): The url rendered or empty string if is None.
        """
        if not self.template.url:
            return ""

        context = Context(self.data)
        template = DjangoTemplate(self.template.url)
        return template.render(context)

    def get_url_msg(self) -> str:
        """
        Renders the url message.
        Returns:
            (str): The url message rendered or empty string if is None.
        """
        if not self.template.url_msg:
            return ""

        context = Context(self.data)
        template = DjangoTemplate(self.template.url_msg)
        return template.render(context)

    def get_recipient(self) -> str:
        """
        Get the recipient string to use it into the provider.

        Returns:
            (str): The recipient for the provider.

        Raises:
            NotificationNotSent if the attribute is not found or is empty
        """
        recipient = getattr(self.user, self.recipient_attr)
        if not recipient:
            raise NotificationNotSent(
                f"Recipient don't have {self.recipient_attr} attribute."
            )
        return recipient

    """
    Status
    """

    def get_callback(self) -> str:
        """ """
        from django.urls import reverse

        return reverse(
            "notifications-callback", args=[self.notification.callback_id]
        )

    @classmethod
    def get_prov_id(cls, request) -> str:
        """
        Gets the provider's id from the response.

        Args:
            request: HTTPResponse after sending the notification to the
            provider.
        """
        raise NotImplementedError()

    @classmethod
    def get_prov_status(cls, request) -> str:
        """
        Gets the provider's status from the response.

        Args:
            request: HTTPResponse after sending the notification to the
            provider.
        """
        # use any provider
        raise NotImplementedError()

    @classmethod
    def update_status(cls, request, callback_id=None):
        """

        Args:
            callback_id: Notification callback_id
            request: HTTPRequest from the provider's callback
        """
        raise NotImplementedError("update status must be implemented.")



from django.db.models import Q
from django.core.exceptions import ObjectDoesNotExist

from swapper import load_model

from .exceptions import (
    WrapperError,
    HandlerDisallowed,
    CategoryDisallowed,
    NotificationNotSent,
)

Template = load_model("txn_notifications", "Template")
Category = load_model("txn_notifications", "Category")


error_msgs = {
    # wrapper
    "no_handlers": "`supported_handlers` is required.",
    "no_active_handlers": "No supported handlers.",
    "no_active_template": "No active templates for this notification.",
    "missing_defaults": "`template_name`, `template_slug` and `title` are "
    "required attributes",
    # user
    "no_active_rcpt": "`recipient` is not an active user.",
    "disallowed_category": "The recipient didn't allowed notifications from "
    "this category.",
    "disallowed_handler": "The recipient didn't allowed notifications from "
    "this handler.",
}


class _TemplateWrapper:
    # template
    template_name = None  # required
    template_description = None
    template_slug = None  # required & unique
    template_category = None

    # default notification template
    title = None  # required
    body = None
    url = None
    url_msg = None
    is_active = True  # default value

    supported_handlers = []

    # ==========================================
    # Wrapper
    # ==========================================
    @classmethod
    def _check_validate_defaults(cls):
        if (
            cls.template_name is None
            or cls.template_slug is None
            or cls.title is None
        ):
            raise WrapperError(
                error_msgs["missing_defaults"], code="missing_defaults"
            )

    @classmethod
    def get_handler(cls, template):
        for handler in cls.supported_handlers:
            if handler.slug == template.handler:
                return handler

        raise WrapperError(
            error_msgs["no_active_handlers"], code="no_active_handlers"
        )

    @classmethod
    def _get_category(cls):
        if cls.template_category:
            return Category.objects.get(slug=cls.template_category)
        return None

    # ==========================================
    # templates
    # ==========================================
    @classmethod
    def create_templates(cls, update_existing=False):
        """
        Creates or update templates.
        It will create a template for every template-handler combination.
        template_slug = f"{template_slug}__{handler.slug}"

        Args:
            update_existing (bool): Must update existing templates?
        """
        cls._check_validate_defaults()

        for handler in cls.supported_handlers:
            template = Template.objects.filter(
                slug=cls.get_template_slug(handler),
                handler=handler.slug,
            ).first()

            if not template:
                cls._create_template(handler)
            elif update_existing:
                cls._update_template(template, handler)

    @classmethod
    def _create_template(cls, handler) -> None:
        """Creates a new template.
        Uses {handler}_{attribute} combination to set the attrbitures or
        default data.

        Args:
            handler (Handler):
        """
        Template.objects.create(
            slug=cls.get_template_slug(handler),
            name=cls.template_name,
            description=cls.template_description,
            handler=handler.slug,
            category=cls._get_category(),
            title=getattr(
                cls,
                f"{handler.slug}_title",
                cls.title,
            ),
            body=getattr(
                cls,
                f"{handler.slug}_body",
                cls.body,
            ),
            url=getattr(
                cls,
                f"{handler.slug}_url",
                cls.url,
            ),
            url_msg=getattr(cls, f"{handler.slug}_url_msg", cls.url_msg),
            is_active=getattr(
                cls,
                f"{handler.slug}_is_active",
                cls.is_active,
            ),
        )

    @classmethod
    def _update_template(cls, template, handler) -> None:
        """
        Updates templates if `update_existing` flag is True on
        `create_templates` function.

        Args:
            template (Template):
            handler (Hander):

        Returns:
            None
        """
        template.name = cls.template_name
        template.description = cls.template_description
        template.category = cls._get_category()

        # uses the handler related attributes or the default attributes.
        template.title = getattr(
            cls,
            f"{handler.slug}_title",
            cls.title,
        )
        template.body = getattr(
            cls,
            f"{handler.slug}_body",
            cls.body,
        )
        template.url = getattr(
            cls,
            f"{handler.slug}_url",
            cls.url,
        )
        template.is_active = getattr(
            cls,
            f"{handler.slug}_is_active",
            cls.is_active,
        )
        template.save()

    @classmethod
    def get_template_slug(cls, handler):
        return f"{cls.template_slug}__{handler.slug}"


class NotificationWrapper(_TemplateWrapper):
    """
    Creates templates and send notifications.
    """

    def __init__(self):
        if not self.supported_handlers:
            raise WrapperError(
                error_msgs["no_handlers"],
                code="no_handlers",
            )

        handlers = [handler.slug for handler in self.supported_handlers]

        self.templates = Template.objects.filter(
            Q(slug__startswith=f"{self.template_slug}__")
            & Q(is_active=True)
            & Q(handler__in=handlers)
            & Q(Q(category__isnull=True) | Q(category__is_active=True))
        )

        if not self.templates.exists():
            raise WrapperError(
                error_msgs["no_active_template"], code="no_active_template"
            )

    def send(
        self, recipient, *, sender=None, target=None, data=None, context=None
    ) -> None:
        """
        Send all the notification using all supported handlers.

        Args:
            recipient (User): Recipient of the notification.
            sender (User): User who sends the notification.
            target: Any django model to wich the notification is related.
            data (dict): Optional. Any data visible to the wrapper and the
                handler. Accesible to the templates.
            context (dict): Optional. Any context visible to the wrapper and
                the handlers

        Raises:
            NotificationNotSent: if the notification is not sent.
        """
        # check that user allow notifications
        self.check_recipient(recipient)

        filtered_handlers = self.get_supported_handlers(
            recipient, sender, target, data, context
        )
        if filtered_handlers != self.supported_handlers:
            templates = self.templates.filter(
                handler__in=[handler.slug for handler in filtered_handlers]
            )
            if not templates:
                raise WrapperError(
                    error_msgs["no_active_template"], code="no_active_template"
                )
        else:
            templates = self.templates.filter(
                handler__in=[
                    handler.slug for handler in self.supported_handlers
                ]
            )

        for template in templates:
            self._send_notification(
                template, recipient, sender=sender, target=target,
                data=data, context=context,
            )

    def send_notification(
        self,
        template_slug,
        recipient,
        *,
        sender=None,
        target=None,
        data=None,
        context=None,
    ):
        """
        Sends a notification using a template, without using send().

        Args:
            template_slug (str): template real slug.
                f.e. {template_slug}__{handler_slug}
            recipient (User):
            sender (User):
            target:
            data (dict):
            context (dict):

        Examples:
            >>> # send the email notfication only
            >>> send_notification("new_announcement__email", *args, **kwargs)
        """
        self.check_recipient(recipient)

        try:
            # expected template__handler format
            template = self.templates.get(slug=template_slug)

        except ObjectDoesNotExist as e:
            raise WrapperError(
                error_msgs["no_active_template"], code="no_active_template"
            ) from e
        self._send_notification(
            template,
            recipient,
            sender=sender,
            target=target,
            data=data,
            context=context,
        )

    def _send_notification(
        self,
        template,
        recipient,
        sender=None,
        target=None,
        data=None,
        context=None,
    ):
        try:
            # check user settings
            self.check_recipient_settings(recipient, template)

            # formatting data
            data = self.get_data(
                data=data,
                recipient=recipient,
                sender=sender,
                target=target,
                context=context,
                template=template,
            )
            HandlerClass = self.get_handler(template)
            handler = HandlerClass(
                recipient, template, sender=sender, data=data
            )
            handler.send()
        except Exception as e:
            self.hadle_exception(e, template, recipient, sender, target, data)

    def get_supported_handlers(self, recipient, sender, target, data, context):
        return self.supported_handlers

    def hadle_exception(
        self,
        exception,
        template,
        recipient,
        sender=None,
        target=None,
        data=None,
    ):
        raise exception

    @classmethod
    def get_data(
        cls,
        *,
        data=None,
        recipient=None,
        sender=None,
        target=None,
        context=None,
        template=None,
        **kwargs,
    ):
        data = data or {}
        data.update(
            {
                "recipient": recipient,
                "sender": sender,
                "target": target,
                "template": template,
                "context": context or {},
            }
        )
        return data

    # ==========================================
    # recipient
    # ==========================================
    @classmethod
    def check_recipient(cls, recipient):
        """

        Args:
            recipient:

        Returns:

        """
        if not recipient.is_active:
            raise NotificationNotSent(
                error_msgs["no_active_rcpt"], code="no_active_rcpt"
            )

    @classmethod
    def check_recipient_settings(cls, user, template):
        """

        Args:
            user (User):
            template (Template):

        Returns:

        """
        print(user, template)

        if user.notification_handler_settings.filter(
            handler=template.handler, allow=False
        ).exists():
            raise HandlerDisallowed(
                error_msgs["disallowed_handler"],
                code="disallowed_handler",
            )

        if template.category:
            if user.notification_category_settings.filter(
                category=template.category,
                allow=False,
            ).exists():
                raise CategoryDisallowed(
                    error_msgs["disallowed_category"],
                    code="disallowed_category",
                )



# Generated by Django 4.1 on 2022-09-02 21:43

from django.conf import settings
from django.db import migrations, models
import django.utils.timezone
import uuid
import swapper


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("description", models.TextField(blank=True, null=True)),
                ("slug", models.SlugField(unique=True)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "created_timestamp",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("updated_timestamp", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Category",
                "verbose_name_plural": "Categories",
                "abstract": False,
                "swappable": swapper.swappable_setting(
                    "txn_notifications", "Category"
                ),
            },
        ),
        migrations.CreateModel(
            name="Template",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("description", models.TextField(blank=True, null=True)),
                ("slug", models.SlugField(unique=True)),
                (
                    "title",
                    models.CharField(
                        blank=True,
                        help_text="Markdown and django template language enabled. You can use {{ <data> }}, for example: {{ recipient }}, {{ sender }}, {{ target }}.",
                        max_length=500,
                        null=True,
                        verbose_name="Notification title",
                    ),
                ),
                (
                    "body",
                    models.TextField(
                        blank=True,
                        help_text="Markdown and django template language enabled. You can use {{ <data> }}, for example: {{ recipient }}, {{ sender }}, {{ target }}.",
                        null=True,
                        verbose_name="Notification body",
                    ),
                ),
                (
                    "url",
                    models.TextField(
                        blank=True,
                        help_text="Markdown and django template language enabled. You can use {{ <data> }}, for example: {{ recipient }}, {{ sender }}, {{ target }}.",
                        max_length=500,
                        null=True,
                        verbose_name="Notification url",
                    ),
                ),
                (
                    "url_msg",
                    models.TextField(
                        blank=True,
                        help_text="Markdown and django template language enabled. You can use {{ <data> }}, for example: {{ recipient }}, {{ sender }}, {{ target }}.",
                        max_length=500,
                        null=True,
                        verbose_name="Notification url message",
                    ),
                ),
                (
                    "handler",
                    models.CharField(help_text="handler slug", max_length=50),
                ),
                (
                    "record",
                    models.BooleanField(
                        default=True,
                        help_text="record the notification in the database using `Notification` model.",
                    ),
                ),
                (
                    "check_status",
                    models.BooleanField(
                        default=True,
                        help_text="check that the notification must update sent status, usually using callbacks or webhooks.",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="if `False`, avoid sending any notification using this template.",
                    ),
                ),
                (
                    "created_timestamp",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                ("updated_timestamp", models.DateTimeField(auto_now=True)),
                (
                    "category",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="templates",
                        to=swapper.get_model_name(
                            "txn_notifications", "Category"
                        ),
                    ),
                ),
            ],
            options={
                "verbose_name": "Template",
                "verbose_name_plural": "Templates",
                "abstract": False,
                "swappable": swapper.swappable_setting(
                    "txn_notifications", "Template"
                ),
            },
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=200)),
                ("body", models.TextField(blank=True, null=True)),
                (
                    "url",
                    models.CharField(blank=True, max_length=500, null=True),
                ),
                (
                    "url_msg",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("data", models.JSONField(null=True)),
                (
                    "read",
                    models.BooleanField(
                        default=False,
                        help_text="Notification has been read for the user?",
                    ),
                ),
                (
                    "read_timestamp",
                    models.DateTimeField(blank=True, null=True),
                ),
                (
                    "sent",
                    models.BooleanField(
                        default=False,
                        help_text="Notification has been confirmed sent to the user?",
                    ),
                ),
                (
                    "sent_timestamp",
                    models.DateTimeField(blank=True, null=True),
                ),
                (
                    "prov_id",
                    models.CharField(
                        blank=True,
                        help_text="Provider's notification id",
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "prov_status",
                    models.CharField(
                        blank=True,
                        help_text="Provider's internal status",
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "target_object_id",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "created_timestamp",
                    models.DateTimeField(default=django.utils.timezone.now),
                ),
                (
                    "callback_id",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sent_notifications",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "target_content_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "template",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notifications",
                        to=swapper.get_model_name(
                            "txn_notifications", "Template"
                        ),
                    ),
                ),
            ],
            options={
                "verbose_name": "Notification",
                "verbose_name_plural": "Notifications",
                "ordering": ("-created_timestamp",),
                "abstract": False,
                "swappable": swapper.swappable_setting(
                    "txn_notifications", "Notification"
                ),
            },
        ),
        migrations.CreateModel(
            name="UserHandlerSetting",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("handler", models.CharField(max_length=50)),
                ("allow", models.BooleanField(default=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notification_handler_settings",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User handler setting",
                "verbose_name_plural": "User handler settings",
                "abstract": False,
                "swappable": swapper.swappable_setting(
                    "txn_notifications", "UserHandlerSetting"
                ),
                "unique_together": {("user", "handler")},
            },
        ),
        migrations.CreateModel(
            name="UserCategorySetting",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("allow", models.BooleanField(default=True)),
                (
                    "category",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=swapper.get_model_name(
                            "txn_notifications", "Category"
                        ),
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="notification_category_settings",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User category setting",
                "verbose_name_plural": "User category settings",
                "abstract": False,
                "swappable": swapper.swappable_setting(
                    "txn_notifications", "UserCategorySetting",
                ),
                "unique_together": {("user", "category")},
            },
        ),
    ]



"""
Templates based on https://github.com/mailgun/transactional-email-templates
"""
from django.conf import settings as django_settings
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string

from txn_notifications.settings import get_settings

from .generic import Handler

settings = get_settings()
email_settings = settings.get("email")
TXT_TEMPLATE = email_settings.get("txt_template")
HTML_TEMPLATE = email_settings.get("html_template")


class EmailHandler(Handler):
    """
    Expected data:
        provided by the wrapper
        ==========================================
        recipient (User): The user who receives the email.
        sender (User): Optional. The user who sends the email.
        template (Template):
        context (dict): Optional.

        email specific
        ==========================================
        from_email (str): Sender's email. If any is set, it will be taken
            from settings.DEFAULT_FROM
        bcc (list): Optional. List of emails.
        cc (list): Optional. List of emails.
        reply_to (list): Optional. List of emails.
        headers (dict): Optional.

        template related
        ==========================================
        context["txt_template"] (str): txt template path
        context["html_template"] (str): html template path

        # template related (using the default template)
        image (str): Optional. An image places before the intro: e.g. A logo.
        image_alt (str): Optional. Image alternative text.
        intro (str): Optional. A message set before the body.
        outro (str): Optional. A message set after the body.
        footer (str): Optional. A message outside the box.

        # any other data
        # ==========================================
        Additional data can be provided, and it will be accessible in the
        template.
        "foo": "bar",
        "x": 100,
    """

    slug = "email"
    recipient_attr = "email"

    check_status = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.email = None

    def pre_send(self) -> None:
        self.email = EmailMultiAlternatives(
            subject=self.title,
            body=self.get_txt_body(),
            from_email=self.get_from_email(),
            to=self.recipient,
            bcc=self.data.get("bcc"),
            connection=self.data.get("connection"),
            attachments=self.data.get("attachments"),
            headers=self.data.get("headers"),
            alternatives=self.data.get("alternatives"),
            cc=self.data.get("cc"),
            reply_to=self.data.get("reply_to"),
        )

        self.email.attach_alternative(self.get_html_body(), "text/html")

    def perform_send(self):
        return self.email.send()

    def get_recipient(self) -> list[str]:
        if type(self.user) == list:
            return [getattr(user, self.recipient_attr) for user in self.user]
        return [getattr(self.user, self.recipient_attr)]

    def get_from_email(self):
        return self.data.get("from_email", django_settings.DEFAULT_FROM_EMAIL)

    def get_txt_body(self):
        template = self.context.get("txt_template", TXT_TEMPLATE)
        return render_to_string(template, self.data)

    def get_html_body(self):
        template = self.context.get("html_template", HTML_TEMPLATE)
        return render_to_string(template, self.data)


class MailgunHandler(EmailHandler):
    @classmethod
    def get_prov_id(cls, request) -> str:
        """

        https://documentation.mailgun.com/en/latest/user_manual.html#sending-via-api
        {
          "message": "Queued. Thank you.",
          "id": "<20111114174239.25659.5817@samples.mailgun.org>"
        }

        Args:
            request:

        Returns:

        """
        data = request.json()
        return data.get("id")

    @classmethod
    def get_prov_status(cls, request) -> str:
        return ""

    @classmethod
    def update_status(cls, request, callback_id=None):
        """
        https://documentation.mailgun.com/en/latest/user_manual.html#webhooks-1

        Args:
            callback_id:
            request:

        Returns:

        """

    @classmethod
    def _verify(cls, signing_key, token, timestamp, signature):
        """
        To verify the webhook is originating from Mailgun you need to:

        * Concatenate timestamp and token values.
        * Encode the resulting string with the HMAC algorithm (using your
            Webhook Signing Key as a key and SHA256 digest mode).
        * Compare the resulting hexdigest to the signature.
        * Optionally, you can cache the token value locally and not honor
            any subsequent request with the same token. This will prevent
            replay attacks.
        * Optionally, you can check if the timestamp is not too far from the
        current time.

        https://www.mailgun.com/blog/email/your-guide-to-webhooks/

        Args:
            signing_key:
            token:
            timestamp:
            signature:

        Returns:

        """
        import hmac
        import hashlib

        hmac_digest = hmac.new(
            key=signing_key.encode(),
            msg=f"{timestamp}{token}".encode(),
            digestmod=hashlib.sha256,
        ).hexdigest()
        return hmac.compare_digest(str(signature), str(hmac_digest))



class NotificationNotSent(Exception):
    """"""

    def __init__(self, msg, code=None):
        self.code = code

        super().__init__(msg)


class WrapperError(NotificationNotSent):
    """For handling any exception raise by the NotificationWrapper"""

    def __init__(self, msg, code=None):
        if not code:
            code = "wrapper_error"
        super().__init__(msg, code)


class Disallowed(NotificationNotSent):
    def __init__(self, msg, code=None):
        if not code:
            code = "disallowed"
        super().__init__(msg, code)


class CategoryDisallowed(Disallowed):
    def __init__(self, msg, code=None):
        if not code:
            code = "category_disallowed"
        super().__init__(msg, code)


class HandlerDisallowed(Disallowed):
    def __init__(self, msg, code=None):
        if not code:
            code = "handler_disallowed"
        super().__init__(msg, code)



from django.urls import path

from .views import callbacks, generic_callbacks

urlpatterns = [
    path(
        "callbacks/<str:handler>/<uuid:callback_id>/",
        callbacks,
        name="notifications-callback",
    ),
    path(
        "callbacks/<str:handler>/",
        generic_callbacks,
        name="notifications-callback-generic",
    ),
]



from django.utils import timezone

from swapper import get_model_name
from dateutil.relativedelta import relativedelta


def remove_old_notifications(time_ago=None):
    Notification = get_model_name("txn_notifications", "Notification")

    if not time_ago:
        time_ago = relativedelta(months=6)
    six_moths_ago = timezone.now() - time_ago
    Notification.objects.filter(created_timestamp__lte=six_moths_ago).delete()



import uuid

from django.db import models
from django.conf import settings
from django.utils import timezone

import swapper

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType


class AbstractCategory(models.Model):
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    slug = models.SlugField(unique=True, db_index=True)

    is_active = models.BooleanField(default=True)

    created_timestamp = models.DateTimeField(default=timezone.now)
    updated_timestamp = models.DateTimeField(auto_now=True)

    class Meta:
        verbose_name = "Category"
        verbose_name_plural = "Categories"
        abstract = True

    def __str__(self):
        return f"[{self.slug}] {self.name}"

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = self.name

        super().save(*args, **kwargs)


class AbstractTemplate(models.Model):
    """
    Multiple templates for multiple handlers

    Templote(slug="welcome__mobile")
    """

    # template data
    name = models.CharField(max_length=100)
    description = models.TextField(null=True, blank=True)
    slug = models.SlugField(unique=True, db_index=True)

    # turn on/off by category
    category = models.ForeignKey(
        swapper.get_model_name("txn_notifications", "Category"),
        on_delete=models.SET_NULL,
        related_name="templates",
        null=True,
    )

    # notification templates / markdown
    title = models.CharField(
        "Notification title",
        max_length=500,
        null=True,
        blank=True,
        help_text="Markdown and django template language enabled. "
        "You can use {{ <data> }}, for example: {{ recipient }}, "
        "{{ sender }}, {{ target }}.",
    )

    body = models.TextField(
        "Notification body",
        null=True,
        blank=True,
        help_text="Markdown and django template language enabled. "
        "You can use {{ <data> }}, for example: {{ recipient }}, "
        "{{ sender }}, {{ target }}.",
    )

    url = models.TextField(
        "Notification url",
        max_length=500,
        null=True,
        blank=True,
        help_text="Markdown and django template language enabled. "
        "You can use {{ <data> }}, for example: {{ recipient }}, "
        "{{ sender }}, {{ target }}.",
    )

    url_msg = models.TextField(
        "Notification url message",
        max_length=500,
        null=True,
        blank=True,
        help_text="Markdown and django template language enabled. "
        "You can use {{ <data> }}, for example: {{ recipient }}, "
        "{{ sender }}, {{ target }}.",
    )

    handler = models.CharField(max_length=50, help_text="handler slug")
    record = models.BooleanField(
        default=True,
        help_text="record the notification in the database using "
        "`Notification` model.",
    )

    check_status = models.BooleanField(
        default=True,
        help_text="check that the notification must update sent status, "
        "usually using callbacks or webhooks.",
    )

    # notifications can be sent using this template
    is_active = models.BooleanField(
        default=True,
        help_text="if `False`, avoid sending any notification using this "
        "template.",
    )

    created_timestamp = models.DateTimeField(default=timezone.now)
    updated_timestamp = models.DateTimeField(auto_now=True)

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = f"{self.name}__{self.handler}"

        if self.check_status:
            self.record = True
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Template"
        verbose_name_plural = "Templates"
        abstract = True

    def __str__(self):
        return f"[{self.slug}] {self.name}"


class NotificationQuerySet(models.query.QuerySet):
    def read(self):
        return self.filter(read=True)

    def unread(self):
        return self.filter(read=False)

    def sent(self):
        return self.filter(sent=True)

    def unsent(self):
        return self.filter(sent=False)

    def mark_all_as_read(self, recipient):
        """Mark as read all notifications for the current recipient."""
        self.unread().filter(recipient=recipient).update(read=True)

    def mark_all_as_unread(self, recipient):
        """Mark as unread all notifications for the current recipient."""
        self.read().filter(recipient=recipient).update(read=False)

    def mark_all_as_sent(self, recipient):
        """Mark as sent all notifications for the current recipient."""
        self.unsent().filter(recipient=recipient).update(sent=True)

    def mark_all_as_unsent(self, recipient):
        """Mark as read unsent all notifications for the current recipient."""
        self.sent().filter(recipient=recipient).update(sent=False)


class AbstractNotification(models.Model):
    recipient = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="notifications",
        on_delete=models.CASCADE,
        blank=False,
        db_index=True,
    )

    sender = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="sent_notifications",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )

    # in case a template have been used
    template = models.ForeignKey(
        swapper.get_model_name("txn_notifications", "Template"),
        on_delete=models.CASCADE,
        related_name="notifications",
    )

    title = models.CharField(max_length=200)
    body = models.TextField(
        null=True,
        blank=True,
    )

    # the main call to action url
    url = models.CharField(
        max_length=500,
        null=True,
        blank=True,
    )
    url_msg = models.CharField(
        max_length=100,
        null=True,
        blank=True,
    )

    # extra context data
    data = models.JSONField(null=True)

    # for handlers that can mark as read
    read = models.BooleanField(
        default=False, help_text="Notification has been read for the user?"
    )
    read_timestamp = models.DateTimeField(null=True, blank=True)

    # for handlers that must check that the notification has been sent using
    # webhooks
    sent = models.BooleanField(
        default=False,
        help_text="Notification has been confirmed sent to the user?",
    )
    sent_timestamp = models.DateTimeField(null=True, blank=True)

    prov_id = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="Provider's notification id",
    )
    prov_status = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        help_text="Provider's internal status",
    )

    # object related to the notification
    target_content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    target_object_id = models.CharField(max_length=255, blank=True, null=True)
    target = GenericForeignKey("target_content_type", "target_object_id")

    created_timestamp = models.DateTimeField(default=timezone.now)

    callback_id = models.UUIDField(default=uuid.uuid4, editable=False)

    objects = NotificationQuerySet.as_manager()

    class Meta:
        verbose_name = "Notification"
        verbose_name_plural = "Notifications"
        abstract = True
        ordering = ("-created_timestamp",)

    def __str__(self):
        return f"{self.recipient} {self.title} {self.timesince()} ago"

    def timesince(self, now=None):
        """
        Shortcut for the ``django.utils.timesince.timesince`` function of the
        current timestamp.
        """
        from django.utils.timesince import timesince as timesince_

        return timesince_(self.created_timestamp, now)

    def mark_as_read(self):
        if not self.read:
            self.read = True
            self.read_timestamp = timezone.now()
            self.save()
        return self.read

    def mark_as_unread(self):
        if self.read:
            self.read = False
            self.save()
        return self.read

    def mark_as_sent(self):
        if not self.sent:
            self.sent = True
            self.sent_timestamp = timezone.now()
            self.save()
        return self.sent


class AbstractUserHandlerSetting(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="notification_handler_settings",
        on_delete=models.CASCADE,
        db_index=True,
    )
    handler = models.CharField(max_length=50)
    allow = models.BooleanField(default=True)

    class Meta:
        verbose_name = "User handler setting"
        verbose_name_plural = "User handler settings"
        unique_together = ("user", "handler")
        abstract = True


class AbstractUserCategorySetting(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="notification_category_settings",
        db_index=True,
    )
    category = models.ForeignKey(
        swapper.get_model_name("txn_notifications", "Category"),
        on_delete=models.CASCADE,
    )
    allow = models.BooleanField(default=True)

    class Meta:
        verbose_name = "User category setting"
        verbose_name_plural = "User category settings"
        unique_together = ("user", "category")
        abstract = True



from swapper import swappable_setting

from txn_notifications.base.models import (
    AbstractCategory,
    AbstractTemplate,
    AbstractNotification,
    AbstractUserHandlerSetting,
    AbstractUserCategorySetting,
)


class Category(AbstractCategory):
    class Meta(AbstractCategory.Meta):
        abstract = False
        swappable = swappable_setting("txn_notifications", "Category")


class Template(AbstractTemplate):
    class Meta(AbstractTemplate.Meta):
        abstract = False
        swappable = swappable_setting("txn_notifications", "Template")


class Notification(AbstractNotification):
    class Meta(AbstractNotification.Meta):
        abstract = False
        swappable = swappable_setting("txn_notifications", "Notification")


class UserCategorySetting(AbstractUserCategorySetting):
    class Meta(AbstractUserCategorySetting.Meta):
        abstract = False
        swappable = swappable_setting(
            "txn_notifications", "UserCategorySetting"
        )


class UserHandlerSetting(AbstractUserHandlerSetting):
    class Meta(AbstractUserHandlerSetting.Meta):
        abstract = False
        swappable = swappable_setting(
            "txn_notifications", "UserHandlerSetting"
        )



from django.test import TestCase

from swapper import load_model

from txn_notifications.tests.factories import (
    UserFactory,
    ReadNotification,
    NotificationFactory,
    SentNotificationFactory,
)

Notification = load_model("txn_notifications", "Notification")


class NotificationQSTestCase(TestCase):
    def setUp(self):
        NotificationFactory()
        NotificationFactory()
        SentNotificationFactory()
        SentNotificationFactory()
        SentNotificationFactory()
        ReadNotification()

    def test_filter_sent(self):
        self.assertEqual(Notification.objects.sent().count(), 3)

    def test_filter_unsent(self):
        self.assertEqual(Notification.objects.unsent().count(), 3)

    def test_filter_unread(self):
        self.assertEqual(Notification.objects.unread().count(), 5)

    def test_filter_read(self):
        self.assertEqual(Notification.objects.read().count(), 1)


class UserNotificationQSTestCase(TestCase):
    def setUp(self):
        self.user = UserFactory()

        NotificationFactory(recipient=self.user)
        NotificationFactory(recipient=self.user)
        SentNotificationFactory(recipient=self.user)
        SentNotificationFactory(recipient=self.user)
        SentNotificationFactory(recipient=self.user)
        ReadNotification(recipient=self.user)

        # other notifications
        NotificationFactory()
        NotificationFactory()
        SentNotificationFactory()
        SentNotificationFactory()
        SentNotificationFactory()
        ReadNotification()

    def test_no_filter(self):
        self.assertEqual(self.user.notifications.count(), 6)

    def test_filter_read(self):
        self.assertEqual(self.user.notifications.read().count(), 1)

    def test_filter_unread(self):
        self.assertEqual(self.user.notifications.unread().count(), 5)

    def test_filter_sent(self):
        self.assertEqual(self.user.notifications.sent().count(), 3)

    def test_filter_unsent(self):
        self.assertEqual(self.user.notifications.unsent().count(), 3)

    def test_mark_all_as_read(self):
        self.assertEqual(self.user.notifications.read().count(), 1)
        Notification.objects.mark_all_as_read(recipient=self.user)
        self.assertEqual(self.user.notifications.unread().count(), 0)

    def test_mark_all_as_unread(self):
        self.assertEqual(self.user.notifications.unread().count(), 5)
        Notification.objects.mark_all_as_unread(recipient=self.user)
        self.assertEqual(self.user.notifications.read().count(), 0)

    def test_mark_all_as_sent(self):
        self.assertEqual(self.user.notifications.sent().count(), 3)
        Notification.objects.mark_all_as_sent(recipient=self.user)
        self.assertEqual(self.user.notifications.unsent().count(), 0)

    def test_mark_all_as_unsent(self):
        self.assertEqual(self.user.notifications.unsent().count(), 3)
        Notification.objects.mark_all_as_unsent(recipient=self.user)
        self.assertEqual(self.user.notifications.sent().count(), 0)



from swapper import load_model

from txn_notifications.base.admin import (
    AbstractCategoryAdmin,
    AbstractTemplateAdmin,
    AbstractNotificationAdmin,
)

from django.contrib import admin

Category = load_model("txn_notifications", "Category")
Template = load_model("txn_notifications", "Template")
Notification = load_model("txn_notifications", "Notification")


@admin.register(Category)
class CategoryAdmin(AbstractCategoryAdmin):
    pass


@admin.register(Template)
class TempalteAdmin(AbstractTemplateAdmin):
    pass


@admin.register(Notification)
class NotificationAdmin(AbstractNotificationAdmin):
    pass



from django.http import HttpResponse
from django.views.decorators.http import require_POST

from txn_notifications.wrappers import NotificationWrapper


@require_POST()
def callbacks(request, handler, callback_id, *args, **kwargs):
    handler = NotificationWrapper.get_handler(handler)
    handler.update_status(request, callback_id)
    return HttpResponse(status=200)


@require_POST()
def generic_callbacks(request, handler, *args, **kwargs):
    handler = NotificationWrapper.get_handler(handler)
    handler.update_status(request)
    return HttpResponse(status=200)



from django.contrib import admin


class AbstractCategoryAdmin(admin.ModelAdmin):
    list_display = ["id", "slug", "name", "is_active"]
    list_filter = ["created_timestamp", "is_active"]
    search_fields = ["slug", "name"]
    readonly_fields = ["created_timestamp"]


class AbstractTemplateAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "slug",
        "name",
        "category",
        "title",
        "is_active",
    )
    list_filter = ["created_timestamp", "is_active"]
    search_fields = ["slug", "name"]
    readonly_fields = ["created_timestamp"]
    autocomplete_fields = ["category"]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("category")


class AbstractNotificationAdmin(admin.ModelAdmin):
    list_display = [
        "id",
        "recipient",
        "title",
        "read",
        "sent",
        "created_timestamp",
    ]
    list_filter = ["created_timestamp", "read", "sent"]
    search_fields = ["recipient__username", "recipient__first_name"]
    readonly_fields = [
        "created_timestamp",
        "recipient",
        "sender",
        "template",
        "target_content_type",
        "target_object_id",
        "sent",
        "sent_timestamp",
        "read",
        "read_timestamp",
    ]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("recipient")



from django.utils import timezone

import factory
import swapper
from factory.django import DjangoModelFactory

from django.contrib.auth import get_user_model

Category = swapper.load_model("txn_notifications", "Category")
Template = swapper.load_model("txn_notifications", "Template")
Notification = swapper.load_model("txn_notifications", "Notification")
UserHandlerSetting = swapper.load_model(
    "txn_notifications", "UserHandlerSetting"
)
UserCategorySetting = swapper.load_model(
    "txn_notifications", "UserCategorySetting"
)


class UserFactory(DjangoModelFactory):
    first_name = factory.Faker("name")
    username = factory.Faker("email")
    email = factory.Faker("email")

    class Meta:
        model = get_user_model()


class CategoryFactory(DjangoModelFactory):
    name = factory.Faker("name")
    description = "category description"
    slug = factory.Faker("lexify", text="category_??????????")

    class Meta:
        model = Category


class TemplateFactory(DjangoModelFactory):
    name = factory.Faker("name")
    description = "template description"
    slug = factory.Faker("lexify", text="template_??????????__django")

    # turn on/off by category
    category = factory.SubFactory(CategoryFactory)

    # notification templates / regex
    title = "hi {{ recipient.first_name }}!"
    body = "this is a **mark_down** template"
    url = "https://app.test.mx/profile/{{ recipient.id }}/"
    url_msg = "See your profile"

    handler = "django"

    class Meta:
        model = Template


class NotificationFactory(DjangoModelFactory):
    recipient = factory.SubFactory(UserFactory)
    sender = factory.SubFactory(UserFactory)

    # in case a template have been used
    template = factory.SubFactory(TemplateFactory)

    title = "notification title"
    body = "notification body"
    url = "notification url"

    target = factory.SubFactory(UserFactory)

    class Meta:
        model = Notification


class SentNotificationFactory(NotificationFactory):
    sent = True
    sent_timestamp = timezone.now()

    prov_id = "123"
    prov_status = "sent"

    class Meta:
        model = Notification


class ReadNotification(NotificationFactory):
    read = True
    read_timestamp = timezone.now()

    class Meta:
        model = Notification


class UserCategorySettingFactory(DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    category = factory.SubFactory(CategoryFactory)
    allow = True

    class Meta:
        model = UserCategorySetting


class UserHandlerSettingFactory(DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    handler = "django"
    allow = True

    class Meta:
        model = UserHandlerSetting



from setuptools import setup

PACKAGE_NAME = "django-transactional-notifications"
PACKAGE_DESCRIPTION = "Send transactional notifications to any service."
PACKAGE_URL = "https://gitlab.com/rogeliomtx/django-notifications"
__version__ = "0.3.3"


setup(
    name=PACKAGE_NAME,
    version=__version__,
    description=PACKAGE_DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Rogelio Martínez",
    author_email="hi@rogermx.com",
    url="https://gitlab.com/rogeliomtx/django-notifications",
    install_requires=[
        "django>=2.2",
        "swapper",
        "markdown",
        "beautifulsoup4",
    ],
    test_requires=[
        "django>=2.2",
        "swapper",
        "markdown",
        "beautifulsoup4",
    ],
    packages=[
        "txn_notifications",
        "txn_notifications.base",
        "txn_notifications.handlers",
        "txn_notifications.templatetags",
        "txn_notifications.templates",
        "txn_notifications.migrations",
        "txn_notifications.tests",
        "txn_notifications.config",
    ],
    package_data={
        "txn_notifications": [
            "templates/txn_notifications/*.html",
            "static/txn_notifications/*.js",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 4.0",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Utilities",
    ],
    keywords="django notifications email twilio mailgun transactional "
    "messages webhooks",
    license="MIT",
)



from django.test import TestCase
from django.utils import timezone

from txn_notifications.tests.factories import NotificationFactory


class NotificationMarkTestCase(TestCase):
    def test_mark_as_read(self):
        notification = NotificationFactory(read=False, read_timestamp=None)

        self.assertFalse(notification.read)
        self.assertIsNone(notification.read_timestamp)

        notification.mark_as_read()

        self.assertTrue(notification.read)
        self.assertIsNotNone(notification.read_timestamp)

    def test_mark_as_unread(self):
        notification = NotificationFactory(
            read=True, read_timestamp=timezone.now()
        )

        self.assertTrue(notification.read)
        self.assertIsNotNone(notification.read_timestamp)

        notification.mark_as_unread()

        self.assertFalse(notification.read)
        self.assertIsNotNone(notification.read_timestamp)

    def test_mark_as_sent(self):
        notification = NotificationFactory(sent=False, sent_timestamp=None)

        self.assertFalse(notification.sent)
        self.assertIsNone(notification.sent_timestamp)

        notification.mark_as_sent()

        self.assertTrue(notification.sent)
        self.assertIsNotNone(notification.sent_timestamp)
